'''PCR reaction(s).'''
import coral


class PrimingError(Exception):
    '''Raise if there is an error during priming (e.g. during PCR).'''


def pcr(template, primer1, primer2, min_tm=50.0, min_primer_len=14):
    '''Simulate a PCR.

    :param template: DNA template from which to PCR.
    :type template: coral.DNA
    :param primer1: First PCR primer.
    :type primer1: coral.Primer
    :param primer2: First PCR primer.
    :type primer2: coral.Primer
    :param min_tm: Minimum melting temperature (Tm) at which primers must bind
                   to the template.
    :type min_tm: float
    :param min_primer_len: Minimum amount of template homology required at the
                           3' end of each primer.
    :type min_primer_len: int
    :returns: A dsDNA Amplicon.
    :rtype: coral.DNA
    :raises: PrimingError if a primer binds more than once on the template,
             primers bind in overlapping sequence of the template, there are no
             forward primer binding sites or reverse priming sites, or if the
             PCR would work only on a circular version of the template (if
             template is linear).

    '''
    # Find match in top or bottom strands for each primer
    p1_matches = coral.analysis.anneal(template, primer1, min_tm=min_tm,
                                       min_len=min_primer_len)
    p2_matches = coral.analysis.anneal(template, primer2, min_tm=min_tm,
                                       min_len=min_primer_len)
    p1_binding_locations = [m[0] for strand in p1_matches for m in strand]
    p2_binding_locations = [m[0] for strand in p2_matches for m in strand]

    # Ensure unique top and bottom matches
    if len(p1_binding_locations) > 1:
        primer_msg = 'Multiple primer 1 binding locations: {}'
        raise PrimingError(primer_msg.format(p1_binding_locations))

    if len(p2_binding_locations) > 1:
        primer_msg = 'Multiple primer 2 binding locations: {}'
        raise PrimingError(primer_msg.format(p2_binding_locations))

    if not p1_binding_locations and not p2_binding_locations:
        raise PrimingError('Neither primer binds the template')

    if not p1_binding_locations:
        raise PrimingError('Primer 1 does not bind the template')

    if not p2_binding_locations:
        raise PrimingError('Primer 2 does not bind the template')

    # Check that primers bind on opposite strands of the template
    tops = p1_matches[0] + p2_matches[0]
    bottoms = p1_matches[1] + p2_matches[1]
    if not tops:
        raise PrimingError('No primers bind the template\'s top strand.')
    if not bottoms:
        raise PrimingError('No primers bind the template\'s bottom strand.')

    # Figure out which primer matches the top strand
    if p1_matches[0]:
        # primer1 is top
        fwd = primer1
        rev = primer2
    else:
        # primer2 matches top strand
        fwd = primer2
        rev = primer1

    # Now we can simulate the PCR. If primer locations are overlapping, we
    # throw an error. If the primers won't amplify a product (e.g. a linear
    # template with primers facing away from one another), throw a different
    # error. Otherwise, amplify the product, including any overhangs.

    # 3' locations, annealing region length
    fwd_3, fwd_len = tops[0]
    rev_3, rev_len = bottoms[0]

    # 5' locations
    fwd_5 = fwd_3 - fwd_len
    rev_5 = rev_3 - rev_len

    # location on the top strand where the 'reverse' primer ends (its 3' end)
    rev_3_top = len(template) - rev_3
    rev_5_top = len(template) - rev_5
    # TODO: Use % operator?
    if rev_5_top > len(template):
        rev_5_top = rev_5_top - len(template)

    # overhangs
    fwd_overhang = fwd.primer()[:-fwd_len].to_ds()
    rev_overhang = rev.primer()[:-rev_len].to_ds()

    # TODO: what about searching substrings over circulate templates?
    # Cases:
    # 1)  Primers point towards one another - overlapping is fine
    #       -> isolate region between 5' annealing regions and tack on the
    #          rest of the overhang.
    # 2)  Primers point away from one another, template is linear
    #       -> error
    # 3)  Primers point away from one another, template is circular
    #       a) Primers don't overlap
    #           -> rotate to 'top' primer start, extract sequence
    #       b) Primers overlap
    #           -> Extract whole sequence as linear fragment, tack on rest of
    #              'bottom' primer. May disrupt features.
    if template.topology == 'circular':
        # Circular template - primers always point towards one another
        if rev_3_top > fwd_3:
            # Inter-primer region doesn't go over the origin (index 0)
            # However, the 'anneal' region may extend over it.
            # FIXME: handle case where 'anneal' region extends over origin
            # FIXME: simplify - just generate 'before' and 'after', then
            # combine with preamplicon later
            if rev_3_top + rev_len > len(template):
                # Reverse primer extends over the origin
                if fwd_5 - fwd_len < 0:
                    # Forward primer extends over the origin
                    preamplicon = template.linearize()
                    # Add extra anneal regions
                    before = template[fwd_5:]
                    after = template[:rev_5_top]
                    preamplicon = before + preamplicon + after
                else:
                    # Only the reverse primer extends over the origin
                    preamplicon = template[fwd_5:]
                    after = template[:rev_5_top]
                    preamplicon = preamplicon + after
            elif fwd_5 - fwd_len < 0:
                # Only the forward primer extends over the origin
                before = template[fwd_5:]
                preamplicon = before + template[:rev_5_top]
            else:
                # Extract like normal
                preamplicon = template[fwd_5:len(template) - rev_5]
        else:
            # Inter-primer region goes over the origin (index 0)
            preamplicon_len = len(template) - fwd_5 + rev_5_top
            preamplicon = template.rotate(fwd_5)[:preamplicon_len]
    else:
        # Linear template
        if rev_3_top < fwd_5 or fwd_3 > rev_5_top:
            # Primers point away from one another.
            raise PrimingError('Primers point away from one another.')
        else:
            # Primers point towards each other.
            preamplicon = template[fwd_5:len(template) - rev_5]

    # Add overhangs
    amplicon = (fwd_overhang.to_ds() +
                preamplicon +
                rev_overhang.to_ds().reverse_complement())

    return amplicon



# -*- coding: utf-8
'''Wrapper for NUPACK 3.0.'''
import multiprocessing
import time
from subprocess import Popen, PIPE
from tempfile import mkdtemp
from shutil import rmtree
from os.path import isdir
from os import environ
from coral.analysis.utils import sequence_type


class Nupack(object):
    '''Run NUPACK functions on sequences.'''
    def __init__(self, seq_list, temp=50, nupack_home=None, rna1999=False):
        '''
        :param seq_list: Input sequence(s).
        :type seq_list: coral.DNA, coral.RNA, or list of
                        either.
        :param rna1999: Use RNA 1999 settings.
        :type rna1999: bool
        :param temp: Temperature in C.
        :type temp: float
        :param nupack_home: NUPACK home dir. The script attempts to find the
                            NUPACKHOME environment variable if this isn't set.
        :type nupack_home: str
        :returns: Nupack instance.
        :raises: ValueError if NUPACKHOME environment variable is not defined
                 and nupack_home is undefined.
                 ValueError if sequences are not all the same type (e.g. don't
                 mix RNA and DNA).
                 Exception if using rna1999 with a Tm other than 37C.

        '''
        # Set up nupack environment variable
        if nupack_home is None:
            if 'NUPACKHOME' in environ:
                nupack_home = environ['NUPACKHOME']
            else:
                msg1 = 'Must have NUPACKHOME environment variable set or '
                msg2 = 'set nupack_home argument.'
                raise ValueError(msg1 + msg2)
        self._nupack_home = nupack_home

        # If input isn't list, make it one
        if type(seq_list) != list:
            self._seq_list = [seq_list]
        else:
            self._seq_list = seq_list

        # Figure out material based on input and ensure it's consistent
        self._material = sequence_type(self._seq_list[0])
        if not all([sequence_type(seq) == self._material for seq in
                   self._seq_list]):
            raise ValueError('Sequence inputs were of mixed types.')

        # Convert seq object(s) to string(s)
        self._seq_list = [str(seq) for seq in self._seq_list]

        # Shared temperature input
        self._temp = temp

        # Handle rna1999 setting
        if self._material == 'rna' and rna1999:
            self._material = 'rna1999'
            if self._temp != 37:
                raise Exception('To use rna1999 settings, temp must be 37.')

        # Create temp dir
        self._tempdir = mkdtemp()

        # Track whether complexes has been run to avoid redundant computation
        self._complexes_run = False
        self._complexes_file = None

    def complexes(self, max_complexes, mfe=True):
        '''Run `complexes` on set of input sequences.

        :param max_complexes: Maximum complex size (integer).
        :type max_complexes: int
        :param mfe: Include mfe calculations (boolean, defaults to True).
        :type mfe: bool
        :returns: complex types (list of interactions) and their energies.
        :rtype: dict

        '''
        self._temp_dir()

        # Prepare input file
        n_seqs = str(len(self._seq_list))
        seqs = '\n'.join(self._seq_list)
        max_complexes = str(max_complexes)
        complexes_input = '\n'.join([n_seqs, seqs, max_complexes])
        with open(self._tempdir + '/nupack.in', 'w') as input_handle:
            input_handle.write(complexes_input)

        # Run 'complexes'
        args = ['-T', str(self._temp), '-material', self._material]
        if mfe:
            args += ['-mfe']
        self._run_cmd('complexes', args)

        # Parse the output
        with open(self._tempdir + '/nupack.cx', 'r+') as output:
            self._complexes_file = output.readlines()
        complexes_results = [x.split() for x in self._complexes_file if '%'
                             not in x]

        # Remove rank entry
        for complexes_result in complexes_results:
            complexes_result.pop(0)

        # Extract complexes
        complexes = []
        for result in complexes_results:
            complex_i = []
            for seq in self._seq_list:
                complex_i.append(int(result.pop(0)))
            complexes.append(complex_i)

        # Extract energies
        energies = [float(x.pop(0)) for x in complexes_results]

        self._complexes_run = int(max_complexes), mfe

        self._close()

        return {'complexes': complexes, 'complex_energy': energies}

    def concentrations(self, max_complexes, conc=[0.5e-6], mfe=True):
        '''Run `concentrations` - get expected steady state concentrations
        of complexes.

        :param max_complexes: Maximum complex size.
        :type max_complexes: int
        :param conc: Oligo concentrations.
        :type conc: list
        :param mfe: Include mfe calculations.
        :type mfe: bool
        :returns: complex types, their concentrations, and their energies.
        :rtype: dict

        '''
        # If complexes hasn't been run, run it
        # TODO: save results of complexes, rewrite to file if necessary.
        # This allows closing tmpdir at every step, makes implementation clean
        if self._complexes_run != (max_complexes, mfe):
            self.complexes(max_complexes=max_complexes, mfe=mfe)

        self._temp_dir()
        with open(self._tempdir + '/nupack.cx', 'w') as cx_file:
            cx_file.writelines(self._complexes_file)

        # Prepare input file
        if len(conc) > 1:
            input_concs = '\n'.join([str(x) for x in conc])
        else:
            input_concs = '\n'.join([str(conc[0]) for x in self._seq_list])
        with open(self._tempdir + '/nupack.con', 'w') as input_handle:
            input_handle.write(input_concs)

        args = ['-sort', str(3)]

        # Run 'concentrations'
        self._run_cmd('concentrations', args)

        # Parse the output of 'complexes'
        with open(self._tempdir + '/nupack.eq', 'r+') as output:
            con_results = output.readlines()
            con_results = [x.split() for x in con_results if '%' not in x]

        # Format results: complex type, concentration, and energy
        # Remove rank information
        for concentration in con_results:
            concentration.pop(0)
        con_types = []
        for result in con_results:
            eq_cx_i = []
            for i in range(len(self._seq_list)):
                eq_cx_i.append(int(float(result.pop(0))))
            con_types.append(eq_cx_i)

        # Extract energies and concentrations
        energies = [x.pop(0) for x in con_results]
        concentrations = [float(x[0]) for x in con_results]

        self._close()

        return {'types': con_types,
                'concentrations': concentrations,
                'energy': energies}

    def mfe(self, index):
        '''Calculate the minimum free energy of a single polymer.

        :param index: Index of strand to analyze.
        :type index: int
        :returns: Minimum Free Energy (mfe).
        :rtype: float

        '''
        self._temp_dir()

        # Prepare input file
        with open(self._tempdir + '/nupack.in', 'w') as input_handle:
            input_handle.write(self._seq_list[index])

        args = ['-T', str(self._temp), '-material', self._material]
        self._run_cmd('mfe', args)
        # Parse the output of 'mfe'

        with open('{}/nupack.mfe'.format(self._tempdir), 'r+') as output:
            mfe = float(output.readlines()[14].strip())

        self._close()

        return mfe

    def pairs(self, index):
        '''Calculate per-pair probability of being unbound for a single
        sequence (secondary structure).

        :param index: Index of strand to analyze.
        :type index: int
        :returns: Unbound pair probability for each base in the sequence.
        :rtype: list

        '''
        self._temp_dir()

        # Sequence at the specified index
        sequence = self._seq_list[index]

        # Prepare input file
        with open(self._tempdir + '/nupack.in', 'w') as input_handle:
            input_handle.write(sequence)

        # Calculate pair probabilities with 'pairs'
        args = ['-T', str(self._temp), '-material', self._material]
        self._run_cmd('pairs', args)

        # Parse the output of 'pairs'
        # Only look at the last n rows - unbound probabilities
        with open(self._tempdir + '/nupack.ppairs', 'r+') as handle:
            pairs = handle.readlines()[-len(sequence):]
            pairs = [pair for pair in pairs if '%' not in pair]

        # Extract pair probability types and pair_probabilities. These are in
        # Nupack's raw text format
        #types = [(int(x.split()[0]), int(x.split()[1])) for x in pairs]
        pair_probabilities = [float(x.split()[2]) for x in pairs]

        self._close()

        return pair_probabilities

    def _close(self):
        '''Delete the temporary dir to prevent filling up /tmp.'''
        rmtree(self._tempdir)

    def _temp_dir(self):
        '''If temporary dir doesn't exist, create it.'''
        if not isdir(self._tempdir):
            self._tempdir = mkdtemp()

    def _run_cmd(self, cmd, cmd_args):
        '''Run NUPACK command line programs.

        :param cmd: NUPACK command line tool to run ('mfe', 'complexes',
                    'concentrations', 'pairs').
        :type cmd: str
        :param cmd_args: Arguments to pass to the command line.
        :type cmd_args: str
        :returns: Variable - whatever `cmd` returns.

        '''
        known_cmds = ['complexes', 'concentrations', 'mfe', 'pairs']
        if cmd not in known_cmds:
            msg = 'Command must be one of: {}'.format(known_cmds)
            raise ValueError(msg)

        cmd_path = '{0}/bin/{1}'.format(self._nupack_home, cmd)
        cmd_args += ['nupack']
        cmd_list = [cmd_path] + cmd_args
        process = Popen(cmd_list,
                        env={'NUPACKHOME': self._nupack_home},
                        cwd=str(self._tempdir), stdout=PIPE)
        process.wait()


def nupack_multi(seqs, material, cmd, arguments, report=True):
    '''Split Nupack commands over processors.

    :param inputs: List of sequences, same format as for coral.analysis.Nupack.
    :type inpus: list
    :param material: Input material: 'dna' or 'rna'.
    :type material: str
    :param cmd: Command: 'mfe', 'pairs', 'complexes', or 'concentrations'.
    :type cmd: str
    :param arguments: Arguments for the command.
    :type arguments: str
    :returns: A list of the same return value you would get from `cmd`.
    :rtype: list

    '''
    nupack_pool = multiprocessing.Pool()
    try:
        args = [{'seq': seq,
                 'cmd': cmd,
                 'material': material,
                 'arguments': arguments} for seq in seqs]
        nupack_iterator = nupack_pool.imap(run_nupack, args)
        total = len(seqs)
        msg = ' calculations complete.'
        passed = 4
        while report:
            completed = nupack_iterator._index
            if (completed == total):
                break
            else:
                if passed >= 4:
                    print '({0}/{1}) '.format(completed, total) + msg
                    passed = 0
                passed += 1
                time.sleep(1)
        multi_output = [x for x in nupack_iterator]
        nupack_pool.close()
        nupack_pool.join()
    except KeyboardInterrupt:
        nupack_pool.terminate()
        nupack_pool.close()
        raise KeyboardInterrupt

    return multi_output


def run_nupack(kwargs):
    '''Run picklable Nupack command.

    :param kwargs: keyword arguments to pass to Nupack as well as 'cmd'.
    :returns: Variable - whatever `cmd` returns.

    '''
    run = Nupack(kwargs['seq'])
    output = getattr(run, kwargs['cmd'])(**kwargs['arguments'])
    return output



'''Primer design tools.'''
import coral
import warnings


def primer(dna, tm=65, min_len=10, tm_undershoot=1, tm_overshoot=3,
           end_gc=False, tm_parameters='cloning', overhang=None,
           structure=False):
    '''Design primer to a nearest-neighbor Tm setpoint.

    :param dna: Sequence for which to design a primer.
    :type dna: coral.DNA
    :param tm: Ideal primer Tm in degrees C.
    :type tm: float
    :param min_len: Minimum primer length.
    :type min_len: int
    :param tm_undershoot: Allowed Tm undershoot.
    :type tm_undershoot: float
    :param tm_overshoot: Allowed Tm overshoot.
    :type tm_overshoot: float
    :param end_gc: Obey the 'end on G or C' rule.
    :type end_gc: bool
    :param tm_parameters: Melting temp calculator method to use.
    :type tm_parameters: string
    :param overhang: Append the primer to this overhang sequence.
    :type overhang: str
    :param structure: Evaluate primer for structure, with warning for high
                      structure.
    :type structure: bool
    :returns: A primer.
    :rtype: coral.Primer
    :raises: ValueError if the input sequence is lower than the Tm settings
             allow.
             ValueError if a primer ending with G or C can't be found given
             the Tm settings.

    '''
    # Check Tm of input sequence to see if it's already too low
    seq_tm = coral.analysis.tm(dna, parameters=tm_parameters)
    if seq_tm < (tm - tm_undershoot):
        msg = 'Input sequence Tm is lower than primer Tm setting'
        raise ValueError(msg)
    # Focus on first 90 bases - shouldn't need more than 90bp to anneal
    dna = dna[0:90]

    # Generate primers from min_len to 'tm' + tm_overshoot
    # TODO: this is a good place for optimization. Only calculate as many
    # primers as are needed. Use binary search.
    primers_tms = []
    last_tm = 0
    bases = min_len
    while last_tm <= tm + tm_overshoot and bases != len(dna):
        next_primer = dna[0:bases]
        last_tm = coral.analysis.tm(next_primer, parameters=tm_parameters)
        primers_tms.append((next_primer, last_tm))
        bases += 1

    # Trim primer list based on tm_undershoot and end_gc
    primers_tms = [(primer, melt) for primer, melt in primers_tms if
                   melt >= tm - tm_undershoot]
    if end_gc:
        primers_tms = [pair for pair in primers_tms if
                       pair[0][-1] == coral.DNA('C') or
                       pair[0][-1] == coral.DNA('G')]
    if not primers_tms:
        raise ValueError('No primers could be generated using these settings')

    # Find the primer closest to the set Tm, make it single stranded
    tm_diffs = [abs(melt - tm) for primer, melt in primers_tms]
    best_index = tm_diffs.index(min(tm_diffs))
    best_primer, best_tm = primers_tms[best_index]
    best_primer = best_primer.to_ss()

    # Apply overhang
    if overhang:
        overhang = overhang.to_ss()

    output_primer = coral.Primer(best_primer, best_tm, overhang=overhang)

    def _structure(primer):
        '''Check annealing sequence for structure.

        :param primer: Primer for which to evaluate structure
        :type primer: sequence.Primer

        '''
        # Check whole primer for high-probability structure, focus in on
        # annealing sequence, report average
        nupack = coral.analysis.Nupack(primer.primer())
        pairs = nupack.pairs(0)
        anneal_len = len(primer.anneal)
        pairs_mean = sum(pairs[-anneal_len:]) / anneal_len
        if pairs_mean < 0.5:
            warnings.warn('High probability structure', Warning)
        return pairs_mean
    if structure:
        _structure(output_primer)
    return output_primer


def primers(dna, tm=65, min_len=10, tm_undershoot=1, tm_overshoot=3,
            end_gc=False, tm_parameters='cloning', overhangs=None,
            structure=False):
    '''Design primers for PCR amplifying any arbitrary sequence.

    :param dna: Input sequence.
    :type dna: coral.DNA
    :param tm: Ideal primer Tm in degrees C.
    :type tm: float
    :param min_len: Minimum primer length.
    :type min_len: int
    :param tm_undershoot: Allowed Tm undershoot.
    :type tm_undershoot: float
    :param tm_overshoot: Allowed Tm overshoot.
    :type tm_overshoot: float
    :param end_gc: Obey the 'end on G or C' rule.
    :type end_gc: bool
    :param tm_parameters: Melting temp calculator method to use.
    :type tm_parameters: string
    :param overhangs: 2-tuple of overhang sequences.
    :type overhangs: tuple
    :param structure: Evaluate each primer for structure, with warning for high
                      structure.
    :type structure: bool
    :returns: A list primers (the output of primer).
    :rtype: list

    '''
    if not overhangs:
        overhangs = [None, None]
    templates = [dna, dna.reverse_complement()]
    primer_list = []
    for template, overhang in zip(templates, overhangs):
        primer_i = primer(template, tm=tm, min_len=min_len,
                          tm_undershoot=tm_undershoot,
                          tm_overshoot=tm_overshoot, end_gc=end_gc,
                          tm_parameters=tm_parameters,
                          overhang=overhang, structure=structure)
        primer_list.append(primer_i)
    return primer_list



'''Coral, core modules for synthetic DNA design.'''
__version__ = '0.3.4'
from . import analysis
from . import constants
from . import database
from . import design
from . import reaction
from . import seqio
from .sequence import DNA
from .sequence import RNA
from .sequence import Peptide
from .sequence import Primer
from .sequence import RestrictionSite
from .sequence import Feature



'''Gibson design module.'''
import coral


class LengthError(Exception):
    '''If primer would be longer than max length, throw this exception'''
    pass


class TmError(Exception):
    '''If the assembly overlap would fall below the minimum Tm, throw this.'''
    pass


def gibson_primers(dna1, dna2, overlap='mixed', maxlen=80, overlap_tm=65.0,
                   insert=None, primer_kwargs=None):
    '''Design Gibson primers given two DNA sequences (connect left to right)

    :param dna1: First piece of DNA for which to design primers. Once Gibsoned,
                 would be connected at its right side to dna2.
    :type dna1: coral.DNA
    :param dna2: First piece of DNA for which to design primers. Once Gibsoned,
                 would be connected at its right side to dna2.
    :type dna2: coral.DNA
    :param overlap: Specifies location of overlap. 'left' puts it on the 'dna1'
                    side (i.e. the primer to amplify dna2). 'right' puts it on
                    the dna2 side, and 'mixed' does a ~50:50 split
    :type overlap: str
    :param maxlen: Maximum length of each primer.
    :type maxlen: int
    :param overlap_tm: Minimum Tm of overlap
    :type overlap_tm: float
    :param insert: A DNA insert to add with primers and use as assembly
                   homology. This overrides the 'split' argument.
    :type insert: coral.DNA
    :param primer_kwargs: keyword arguments to pass to design_primer()
    :type primer_kwargs: dict
    :returns: Reverse, then forward primer for bridging the two sequences.
              Note that the forward primer binds dna2, reverse dna1.
    :rtype: A sequence.Primer tuple
    :raises: ValueError if split parameter is an invalid string.

    '''
    if primer_kwargs is None:
        primer_kwargs = {}

    # Annealing sequences
    # DNA 2 primer is a forward primer
    fwd_anneal = coral.design.primer(dna2, **primer_kwargs)
    # DNA 1 primer is a reverse primer
    rev_anneal = coral.design.primer(dna1.flip(), **primer_kwargs)
    # Overhangs
    if insert is None:
        # No insert, so follow split argument
        if overlap == 'left':
            # If splitting left, put overhang on forward primer
            overlap_revcomp = coral.design.primer(dna1.flip(), tm=overlap_tm,
                                                  tm_undershoot=0)
            fwd_overhang = overlap_revcomp.primer().reverse_complement()
            rev_overhang = None
        elif overlap == 'right':
            # If splitting right, put overhang on reverse primer
            overlap = coral.design.primer(dna2, tm=overlap_tm, tm_undershoot=0)
            fwd_overhang = None
            rev_overhang = overlap.primer().reverse_complement()
        elif overlap == 'mixed':
            # If mixed, grow size of both until overlap Tm is reached
            overlap_l = dna1[0:0]  # Empty sequence.DNA
            overlap_r = dna2[0]  # First base
            overlap_melt = overlap_r.tm()
            while overlap_melt < overlap_tm:
                rlen = len(overlap_r)
                llen = len(overlap_l)
                if rlen > llen:
                    # Increase left side of overlap
                    overlap_l = dna1[-(rlen + 1):]
                else:
                    # Increase right side of overlap
                    overlap_r = dna2[:(llen + 1)]
                overlap = overlap_l + overlap_r
                overlap_melt = overlap.tm()
            fwd_overhang = overlap_l
            rev_overhang = overlap_r.reverse_complement()
        else:
            raise ValueError('split argument must be left, right, or mixed')
        # Generate primers using anneal, overhang, and tm data
        fwd = coral.Primer(fwd_anneal.primer(), tm=fwd_anneal.tm,
                           overhang=fwd_overhang)
        rev = coral.Primer(rev_anneal.primer(), tm=rev_anneal.tm,
                           overhang=rev_overhang)
    else:
        # There's an insert to use as the overhang
        overlap = insert
        fwd_overhang = insert.to_ss()
        rev_overhang = insert.reverse_complement().to_ss()
        # Generate primers using anneal, overhang, and tm data
        fwd = coral.Primer(fwd_anneal.primer(), tm=fwd_anneal.tm,
                           overhang=fwd_overhang)
        rev = coral.Primer(rev_anneal.primer(), tm=rev_anneal.tm,
                           overhang=rev_overhang)
        left_trim = 0
        # If either primer is too long, try trimming the overhang
        while len(fwd) > maxlen:
            # Generate new overlap
            overlap = insert[left_trim:]
            # Tm must be above overlap_tm
            if coral.analysis.tm(overlap) < overlap_tm:
                raise TmError('Right primer is too long with this Tm setting.')
            # Regenerate forward overhang
            fwd_overhang = overlap.to_ss()
            # Regenerate primer with new overhang
            fwd = coral.Primer(fwd_anneal.primer(), tm=fwd_anneal.tm,
                               overhang=fwd_overhang)
            # Increase 'trimming' index
            left_trim += 1
        right_trim = 0
        while len(rev) > maxlen:
            # Generate new overlap
            overlap = insert[:len(insert) - right_trim]
            # Tm must be above overlap_tm
            if coral.analysis.tm(overlap) < overlap_tm:
                raise TmError('Left primer is too long with this Tm setting.')
            # Regenerate reverse overhang
            rev_overhang = overlap.reverse_complement().to_ss()
            rev = coral.Primer(rev_anneal.primer(), tm=rev_anneal.tm,
                               overhang=rev_overhang)
            # Increase 'trimming' index
            right_trim += 1
    # Check primer lengths
    if any([len(primer) > maxlen for primer in (fwd, rev)]):
        raise LengthError('At least one of the primers is longer than maxlen.')

    return rev, fwd


def gibson(seq_list, circular=True, overlaps='mixed', overlap_tm=65,
           maxlen=80, terminal_primers=True, primer_kwargs=None):
    '''Design Gibson primers given a set of sequences

    :param seq_list: List of DNA sequences to stitch together
    :type seq_list: list containing coral.DNA
    :param circular: If true, designs primers for making a circular construct.
                     If false, designs primers for a linear construct.
    :type circular: bool
    :param splits: Specifies locations of overlap. Must be either a single
                   entry of the same type as the 'split' parameter in
                   gibson_primers or a list of those types of the appropriate
                   length (for circular construct, len(seq_list), for
                   linear construct, len(seq_list) - 1)
    :type splits: str or list of str
    :param overlap_tm: Minimum Tm of overlap
    :type overlap_tm: float
    :param maxlen: Maximum length of each primer.
    :type maxlen: int
    :param terminal_primers: If the output is not circular, will design
                             non-Gibson primers for amplifying the first and
                             last fragments sans homology. If False, there will
                             be one less set of primers returned.
    :type terminal_primers: bool
    :param primer_kwargs: keyword arguments to pass to design.primer
    :type primer_kwargs: dict
    :returns: Forward and reverse primers for amplifying every fragment.
    :rtype: a list of sequence.Primer tuples
    :raises: ValueError if split parameter is an invalid string or wrong size.

    '''

    # Input checking
    if circular:
        n_overlaps = len(seq_list)
    else:
        n_overlaps = len(seq_list) - 1

    if type(overlaps) is str:
        overlaps = [overlaps] * n_overlaps
    else:
        if len(overlaps) != n_overlaps:
            raise ValueError('Incorrect number of \'overlaps\' entries.')
        else:
            for overlap in overlaps:
                if overlap not in ['left', 'right', 'mixed']:
                    raise ValueError('Invalid \'overlaps\' setting.')

    if primer_kwargs is None:
        primer_kwargs = {}

    # If here, inputs were good
    # Design primers for linear constructs:
    primers_list = []
    for i, (left, right) in enumerate(zip(seq_list[:-1], seq_list[1:])):
        primers_list.append(gibson_primers(left, right, overlaps[i],
                                           overlap_tm=overlap_tm,
                                           primer_kwargs=primer_kwargs))
    if circular:
        primers_list.append(gibson_primers(seq_list[-1], seq_list[0],
                                           overlaps[-1],
                                           overlap_tm=overlap_tm,
                                           primer_kwargs=primer_kwargs))
    else:
        if terminal_primers:
            primer_f = coral.design.primer(seq_list[0], **primer_kwargs)
            primer_r = coral.design.primer(seq_list[-1].reverse_complement(),
                                           **primer_kwargs)
            primers_list.append((primer_r, primer_f))

    # Primers are now in order of 'reverse for seq1, forward for seq2' config
    # Should be in 'forward and reverse primers for seq1, then seq2', etc
    # Just need to rotate one to the right
    flat = [y for x in primers_list for y in x]
    flat = [flat[-1]] + flat[:-1]
    grouped_primers = [(flat[2 * i], flat[2 * i + 1]) for i in
                       range(len(flat) / 2)]

    return grouped_primers


def gibson_equimolar(lengths, concs):
    pass



'''Needleman-Wunsch alignment functions.'''
import multiprocessing
import coral
try:
    from .calign import aligner, score_alignment
except ImportError:
    from .align import aligner, score_alignment


def needle(reference, query, gap_open=-15, gap_extend=0, matrix='DNA_simple'):
    '''Do a Needleman-Wunsch alignment.

    :param reference: Reference sequence.
    :type reference: coral.DNA
    :param query: Sequence to align against the reference.
    :type query: coral.DNA
    :param gapopen: Penalty for opening a gap.
    :type gapopen: float
    :param gapextend: Penalty for extending a gap.
    :type gapextend: float
    :param matrix: Matrix to use for alignment - options are DNA_simple (for
                   DNA) and BLOSUM62 (for proteins).
    :type matrix: str
    :returns: (aligned reference, aligned query, score)
    :rtype: tuple of two coral.DNA instances and a float

    '''
    # Align using cython Needleman-Wunsch
    aligned_ref, aligned_res = aligner(str(reference),
                                       str(query),
                                       gap_open=gap_open,
                                       gap_extend=gap_extend,
                                       method='global_cfe',
                                       matrix=matrix)

    # Score the alignment
    score = score_alignment(aligned_ref, aligned_res, gap_open, gap_extend,
                            'DNA_simple')

    return coral.DNA(aligned_ref), coral.DNA(aligned_res), score


def run_needle(args):
    '''Run needle command using 4-tuple of the arguments (in the same order)
    as is used for needle. Necessary to make picklable function for
    multiprocessing.'''
    return needle(*args)


def needle_multi(references, queries, gap_open=-15, gap_extend=0,
                 matrix='DNA_simple'):
    '''Batch process of sequencing split over several cores. Acts just like
    needle but sequence inputs are lists.

    :param references: References sequence.
    :type references: coral.DNA list
    :param queries: Sequences to align against the reference.
    :type queries: coral.DNA list
    :param gap_open: Penalty for opening a gap.
    :type gap_open: float
    :param gap_extend: Penalty for extending a gap.
    :type gap_extend: float
    :param matrix: Matrix to use for alignment - options are DNA_simple (for
                   DNA) and BLOSUM62 (for proteins).
    :type matrix: str
    :returns: a list of the same output as coral.sequence.needle
    :rtype: list

    '''
    pool = multiprocessing.Pool()
    try:
        args_list = [[ref, que, gap_open, gap_extend, matrix] for ref, que in
                     zip(references, queries)]
        aligned = pool.map(run_needle, args_list)
    except KeyboardInterrupt:
        print 'Caught KeyboardInterrupt, terminating workers'
        pool.terminate()
        pool.join()
        raise KeyboardInterrupt

    return aligned



import os
from tempfile import mkdtemp
from Bio import Entrez
import coral.sequence


# FIXME: If a Genome is a data structure, it should be in the DNA sequence
# module. Then rename the genome acquirer after Entrez / NCBI, etc.
# FIXME: If a Genome is not a new special data structure, should be a function
# not a class unless there are more operations than fetch.
# TODO: docstring
# TODO: Figure out why reading in the DNA is so slow and if it can be sped up
# - MG1655 takes 30-60 seconds to process into memory and coral.DNA.
# MG1655 id is 'U00096.3'
def fetch_genome(genome_id):
    '''Acquire a genome from Entrez

    '''
    # TODO: Can strandedness by found in fetched genome attributes?
    # TODO: skip read/write step?
    # Using a dummy email for now - does this violate NCBI guidelines?
    email = 'loremipsum@gmail.com'
    Entrez.email = email

    print 'Downloading Genome...'
    handle = Entrez.efetch(db='nucleotide', id=str(genome_id), rettype='gb',
                           retmode='text')
    print 'Genome Downloaded...'
    tmpfile = os.path.join(mkdtemp(), 'tmp.gb')
    with open(tmpfile, 'w') as f:
        f.write(handle.read())
    genome = coral.seqio.read_dna(tmpfile)

    return genome



'''Cloning design classes and functions.'''
from ._primers import primer, primers
from ._sequence_generation.random_sequences import random_dna
from ._sequence_generation.random_sequences import random_codons
from ._oligo_synthesis.oligo_assembly import OligoAssembly
from ._gibson import gibson_primers
from ._gibson import gibson



'''Analyze sequences.'''
from ._sequence.anneal import anneal
from ._sequence.melting_temp import tm
from ._sequence.repeats import repeats
from ._sequencing.needle import needle
from ._sequencing.needle import needle_multi
from ._sequencing.sanger import Sanger
from ._structure.nupack import Nupack
from ._structure.nupack import nupack_multi
from ._structure.dimers import dimers
from ._structure.vienna import Vienna
from ._structure.structure_windows import StructureWindows
from . import utils



'''Primer Annealing Event Simulation.'''


class PrimerLengthError(Exception):
    '''Primer does not meet minimum length requirements.'''


class AnnealError(Exception):
    '''Generic primer annealing error.'''


def anneal(template, primer, min_tm=50.0, min_len=10):
    '''Simulates a primer binding event. Will find the maximum subset
    of bases in the primer that binds to the template, including overhang
    sequences. **Note**: Primer binding locations indicate the 3' end of the
    primer, not the begining of the annealing sequence.

    :param template: DNA template for which to bind a primer.
    :type template: coral.DNA
    :param primer: Primer to bind to template.
    :type primer: coral.Primer
    :param min_tm: The cutoff melting temperature for primer binding - a binder
                   with a lower Tm will be rejected.
    :type min_tm: float
    :param min_len: The cutoff for bases required for binding - a binder with
                    fewer bases will be rejected.
    :type min_len: int
    :returns: A length 2 list (top and bottom strands) of matches. Each
              match is itself a 2-tuple indicating (1) the location on the
              template of the 3' end of the primer binding site and (2) the
              length of the match (number of bases), e.g. [[(25, 15)],[]] would
              indicate a single top-strand match at template position 25 with
              15 bases of 3' primer homology.
    :rtype: list
    :raises: PrimerLengthError if primer length is too small.
             AnnealError if inputs are of the wrong type.

    '''
    # TODO: add possibility for primer basepair mismatch
    if len(primer) < min_len:
        msg = 'Primer length is shorter than min_len argument.'
        raise PrimerLengthError(msg)
    if len(template) < min_len:
        msg = 'Template is shorter than the min_len argument.'
        raise AnnealError(msg)

    # Strategy: locate all min-length matches, then extend them until they
    # no longer match. This provides an advantage over the previous strategy of
    # updating a dictionary with indices from coral.DNA.locate() as keys, as
    # the latter's indices may actually move for a given primer as it passes
    # over the origin
    def update_match_linear(base, location_length, anneal_seq):
        '''Increase the location and length of binding site, if applicable.'''
        # TODO: this is very inefficient - should stop updating once the first
        # mismatch occurs.
        location, length = location_length

        if location == 0:
            return location_length

        location_next = location - 1
        length_next = length + 1
        seq = base[location_next:location_next + length_next]

        if seq == anneal_seq:
            return (location_next, length_next)
        else:
            return location_length

    def update_match_circular(base, location_length, anneal_seq):
        '''Increase the location and length of binding site, if applicable.'''
        # TODO: this is very inefficient - should stop updating once the first
        # mismatch occurs.
        base_len = len(base)
        location, length = location_length
        if location == 0:
            location_next = base_len - 1
        else:
            location_next = location - 1
        length_next = length + 1

        if (location_next + length_next) > base_len:
            upstream = base[location_next:]
            downstream = base[:length_next - (base_len - location_next)]
            seq = upstream + downstream
        else:
            # No need to 'rotate' sequence
            seq = base[location_next:location_next + length_next]

        if seq == anneal_seq:
            return (location_next, length_next)
        else:
            return location_length

    if template.topology == 'circular':
        update_fun = update_match_circular
    else:
        update_fun = update_match_linear

    # Maximum annealing length to test (can't exceed template length)
    max_len = min(len(template), len(primer))

    primer_dna = primer.primer().to_ds()
    anneal_len = min_len
    anneal_seq = primer_dna[-anneal_len:]
    binding_data = []
    for k, strand_locs in enumerate(template.locate(anneal_seq)):
        matches = zip(strand_locs, [min_len] * len(strand_locs))
        for i in range(anneal_len + 1, max_len + 1):
            anneal_seq = primer_dna[-i:]
            for j, match in enumerate(matches):
                if k == 0:
                    matches[j] = update_fun(template.top(), match, anneal_seq)
                else:
                    matches[j] = update_fun(template.bottom(), match,
                                            anneal_seq)
        binding_data.append(matches)

    # Now, filter out all the matches that are too short
    for i in reversed(range(len(primer_dna) + 1)):
        min_len = i + 1
        tm = primer_dna[-min_len:].tm()
        if tm < min_tm:
            break

    for strand in binding_data:
        for i in reversed(range(len(strand))):
            if strand[i][1] < min_len:
                strand.pop(i)

    # Finally, adjust the position to be the 3' end
    for strand in binding_data:
        for i, match in enumerate(strand):
            length = match[1]
            loc_new = match[0] + length
            if loc_new > len(template):
                # Circularly permute
                loc_new = loc_new - len(template)
            strand[i] = [loc_new, length]

    # Overwriting dictionary keys ensures uniqueness
    # fwd_matches = {}
    # rev_matches = {}
    # for i in range(len(primer) - min_len + 1)[::-1]:
    #     primer_dna = primer.overhang + primer.anneal
    #     annealing = primer_dna[i:]
    #     anneal_temp = annealing.tm()
    #     anneal_len = len(annealing)
    #     if anneal_temp > min_tm:
    #         p_matches = template.locate(annealing)
    #         for match in p_matches[0]:
    #             fwd_matches[match + anneal_len] = anneal_len
    #         for match in p_matches[1]:
    #             rev_matches[match + anneal_len] = anneal_len

    # # Convert dictionaries to lists
    # fwds = [[key, val] for key, val in fwd_matches.iteritems()]
    # revs = [[key, val] for key, val in rev_matches.iteritems()]

    return binding_data



'''Check for primer dimers using Nupack.'''
import coral.analysis


def dimers(primer1, primer2, concentrations=[5e-7, 3e-11]):
    '''Calculate expected fraction of primer dimers.

    :param primer1: Forward primer.
    :type primer1: coral.DNA
    :param primer2: Reverse primer.
    :type primer2: coral.DNA
    :param template: DNA template.
    :type template: coral.DNA
    :param concentrations: list of concentrations for primers and the
                           template. Defaults are those for PCR with 1kb
                           template.
    :type concentrations: list
    :returns: Fraction of dimers versus the total amount of primer added.
    :rtype: float

    '''
    # It is not reasonable (yet) to use a long template for doing these
    # computations directly, as NUPACK does an exhaustive calculation and
    # would take too long without a cluster.
    # Instead, this function compares primer-primer binding to
    # primer-complement binding

    # Simulate binding of template vs. primers
    nupack = coral.analysis.Nupack([primer1.primer(), primer2.primer(),
                                    primer1.primer().reverse_complement(),
                                    primer2.primer().reverse_complement()])
    # Include reverse complement concentration
    primer_concs = [concentrations[0]] * 2
    template_concs = [concentrations[1]] * 2
    concs = primer_concs + template_concs
    nupack_concs = nupack.concentrations(2, conc=concs)
    dimer_conc = nupack_concs['concentrations'][5]
    #primer1_template = nupack_concs['concentrations'][6]
    #primer2_template = nupack_concs['concentrations'][10]
    return dimer_conc / concs[0]



'''RNA module.'''
import coral.reaction
from ._sequence import Sequence
from ._sequence import reverse_complement


class RNA(object):
    '''RNA sequence.'''
    def __init__(self, rna, features=None, run_checks=True):
        '''
        :param rna: Input sequence (RNA).
        :type rna: str
        :param run_checks: Check inputs / formats (disabling increases speed):
                           alphabet check
                           case
        :type run_checks: bool
        :returns: coral.RNA instance.

        '''
        self._sequence = Sequence(rna, 'rna', run_checks=run_checks,
                                  any_char='N')
        if features is None:
            self.features = []
        else:
            self.features = features

    def copy(self):
        '''Create a copy of the current instance.

        :returns: A safely-editable copy of the current sequence.
        :rtype: coral.RNA

        '''
        return type(self)(self._sequence.seq, features=self.features,
                          run_checks=False)

    def locate(self, pattern):
        return self._sequence.locate(pattern)

    def reverse_complement(self):
        '''Reverse complement sequence.

        :returns: The reverse-complement of the current sequence.
        :rtype: coral.RNA

        '''
        new_instance = self.copy()
        # Overwrite sequence with reverse complement
        new_instance[:] = reverse_complement(self._sequence, 'rna')

        return new_instance

    def extract(self, feature, remove_subfeatures=False):
        return self._sequence.extract(feature,
                                      remove_subfeatures=remove_subfeatures)

    def reverse_transcribe(self):
        '''Reverse transcribe to DNA.

        :returns: The reverse transcribed (DNA) version of the current RNA.
        :rtype: coral.DNA

        '''
        return coral.reaction.reverse_transcribe(self)

    def translate(self):
        '''Translate sequence into a peptide.

        :returns: A translated peptide from the current sequence.
        :rtype: coral.Peptide

        '''
        return coral.reaction.translate(self)

    def __add__(self, other):
        # Merge the sequences
        copy = self.copy()
        ocopy = other.copy()
        copy._sequence += ocopy._sequence

        # Merge the features
        for feature in ocopy.features:
            feature.move(len(copy))
        copy.features += ocopy.features

        return copy

    def __contains__(self, key):
        return key in self._sequence

    def __delitem__(self, key):
        self._sequence.__delitem__(key)

    def __eq__(self, other):
        return self._sequence == other._sequence

    def __getitem__(self, key):
        new_sequence = self._sequence[key]
        new_instance = type(self)(str(new_sequence), features=self.features,
                                  run_checks=False)
        return new_instance

    def __len__(self):
        return len(self._sequence)

    def __mul__(self, n):
        # TODO: Keep features as well?
        copy = self.copy()
        copy._sequence = self._sequence.__mul__(n)
        return copy

    def __ne__(self, other):
        return self._sequence != other._sequence

    def __radd__(self, other):
        if other == 0 or other is None:
            # For compatibility with sum()
            return self
        else:
            copy = self.copy()
            try:
                return copy._sequence.__radd__(other._sequence)
            except AttributeError:
                raise TypeError('Cannot add {} to {}'.format(self, other))

    def __repr__(self):
        '''String to print when object is called directly.'''
        header = 'RNA:'
        sequence = self._sequence.__repr__()
        return ' '.join([header, sequence])

    def __setitem__(self, key, value):
        self._sequence.__setitem__(key, value)

    def __str__(self):
        return str(self._sequence)



'''Peptide module.'''
from ._sequence import Sequence


class Peptide(object):
    '''Peptide sequence.'''
    def __init__(self, peptide, features=None, run_checks=True):
        '''
        :param peptide: Input sequence (peptide).
        :type peptide: str
        :param run_checks: Check inputs / formats (disabling increases speed):
                           alphabet check
                           case
        :type run_checks: bool
        :returns: coral.Peptide instance.

        '''
        self._sequence = Sequence(peptide, 'peptide', run_checks=run_checks,
                                  any_char='X')
        if features is None:
            self.features = []
        else:
            self.features = features

    def copy(self):
        '''Create a copy of the current instance.

        :returns: A safely editable copy of the current sequence.
        :rtype: coral.Peptide

        '''
        return type(self)(str(self._sequence), features=self.features,
                          run_checks=False)

    @classmethod
    def extract(self, feature, remove_subfeatures=False):
        return self._sequence.extract(self, feature,
                                      remove_subfeatures=remove_subfeatures)

    def locate(self, pattern):
        return self._sequence.locate(pattern)

    def __add__(self, other):
        # Merge the sequences
        copy = self.copy()
        ocopy = other.copy()
        copy._sequence += ocopy._sequence

        # Merge the features
        for feature in ocopy.features:
            feature.move(len(copy))
        copy.features += ocopy.features

        return copy

    def __contains__(self, key):
        return key in self._sequence

    def __delitem__(self, key):
        self._sequence.__delitem__(key)

    def __eq__(self, other):
        return self._sequence == other._sequence

    def __getitem__(self, key):
        new_instance = type(self)(str(self._sequence[key]),
                                  features=self.features, run_checks=False)
        return new_instance

    def __len__(self):
        return len(self._sequence)

    def __mul__(self, n):
        # TODO: Keep features as well?
        copy = self.copy()
        copy._sequence = self._sequence.__mul__(n)
        return copy

    def __ne__(self, other):
        return self._sequence != other._sequence

    def __radd__(self, other):
        if other == 0 or other is None:
            # For compatibility with sum()
            return self
        else:
            copy = self.copy()
            try:
                return copy._sequence.__radd__(other._sequence)
            except AttributeError:
                raise TypeError('Cannot add {} to {}'.format(self, other))

    def __repr__(self):
        '''String to print when object is called directly.'''
        header = 'Peptide:'
        sequence = self._sequence.__repr__()
        return ' '.join([header, sequence])

    def __setitem__(self, key, value):
        self._sequence.__setitem__(key, value)

    def __str__(self):
        return str(self._sequence)



'''Read and write cloning-relevant formats (sequences, chromatograms, etc).'''
from coral.seqio._dna import read_dna
from coral.seqio._dna import read_sequencing
from coral.seqio._dna import write_dna
from coral.seqio._dna import write_primers



'''Generate overlapping oligo sequences to assemble a larger DNA sequence.'''
import csv
import coral


class OligoAssembly(object):
    '''Split a sequence into overlapping oligonucleotides.'''
    def __init__(self, dna, tm=72, length_range=(80, 200), require_even=True,
                 start_5=True, oligo_number=None, overlap_min=20,
                 min_exception=False):
        '''
        :param dna: Sequence to split into overlapping oligos.
        :type dna: coral.DNA
        :param primers: Design cloning primers that bind the termini of the
                        input sequence.
        :type primers: bool
        :param primer_tm: Ideal Tm for the cloning primers, if applicable.
        :type primer_tm: float
        :param length_range: Maximum oligo size (e.g. 60bp price point cutoff)
                             range - lower bound only matters if oligo_number
                             parameter is set.
        :type length_range: int 2-tuple
        :param require_even: Require that the number of oligonucleotides is
                             even.
        :type require_even: bool
        :param start_5: Require that the first oligo's terminal side is 5\'.
        :type start_5: bool
        :param oligo_number: Attempt to build assembly with this many oligos,
                             starting with length_range min and incrementing by
                             10 up to length_range max.
        :type oligo_number: bool
        :param overlap_min: Minimum overlap size.
        :type overlap_min: int
        :param min_exception: In order to meet tm and overlap_min requirements,
                              allow overlaps less than overlap_min to continue
                              growing above tm setpoint.
        :type min_exception: bool
        :returns: coral.design.OligoAssembly instance.

        '''
        self.kwargs = {'tm': tm, 'length_range': length_range,
                       'require_even': require_even, 'start_5': start_5,
                       'oligo_number': oligo_number,
                       'overlap_min': overlap_min,
                       'min_exception': min_exception}
        self.template = dna
        self.oligos = None
        self.overlaps = None
        self.overlap_tms = None
        self.primers = None
        self.overlap_indices = None

        self._has_run = False
        self.warning = None

    def design_assembly(self):
        '''Design the overlapping oligos.

        :returns: Assembly oligos, and the sequences, Tms, and indices of their
                  overlapping regions.
        :rtype: dict

        '''
        # Input parameters needed to design the oligos
        length_range = self.kwargs['length_range']
        oligo_number = self.kwargs['oligo_number']
        require_even = self.kwargs['require_even']
        melting_temp = self.kwargs['tm']
        overlap_min = self.kwargs['overlap_min']
        min_exception = self.kwargs['min_exception']
        start_5 = self.kwargs['start_5']

        if len(self.template) < length_range[0]:
            # If sequence can be built with just two oligos, do that
            oligos = [self.template, self.template.reverse_complement()]
            overlaps = [self.template]
            overlap_tms = [coral.analysis.tm(self.template)]
            assembly_dict = {'oligos': oligos, 'overlaps': overlaps,
                             'overlap_tms': overlap_tms}

            self.oligos = assembly_dict['oligos']
            self.overlaps = assembly_dict['overlaps']
            self.overlap_tms = assembly_dict['overlap_tms']
            return assembly_dict

        if oligo_number:
            # Make first attempt using length_range[1] and see what happens
            step = 3  # Decrease max range by this amount each iteration

            length_max = length_range[1]
            current_oligo_n = oligo_number + 1
            oligo_n_met = False
            above_min_len = length_max > length_range[0]
            if oligo_n_met or not above_min_len:
                raise Exception('Failed to design assembly.')
            while not oligo_n_met and above_min_len:
                # Starting with low range and going up doesnt work for longer
                # sequence (overlaps become longer than 80)
                assembly = _grow_overlaps(self.template, melting_temp,
                                          require_even, length_max,
                                          overlap_min, min_exception)
                current_oligo_n = len(assembly[0])
                if current_oligo_n > oligo_number:
                    break
                length_max -= step
                oligo_n_met = current_oligo_n == oligo_number
        else:
            assembly = _grow_overlaps(self.template, melting_temp,
                                      require_even, length_range[1],
                                      overlap_min, min_exception)

        oligos, overlaps, overlap_tms, overlap_indices = assembly

        if start_5:
            for i in [x for x in range(len(oligos)) if x % 2 == 1]:
                oligos[i] = oligos[i].reverse_complement()
        else:
            for i in [x for x in range(len(oligos)) if x % 2 == 0]:
                oligos[i] = oligos[i].reverse_complement()

        # Make single-stranded
        oligos = [oligo.to_ss() for oligo in oligos]

        assembly_dict = {'oligos': oligos,
                         'overlaps': overlaps,
                         'overlap_tms': overlap_tms,
                         'overlap_indices': overlap_indices}

        self.oligos = assembly_dict['oligos']
        self.overlaps = assembly_dict['overlaps']
        self.overlap_tms = assembly_dict['overlap_tms']
        self.overlap_indices = assembly_dict['overlap_indices']

        for i in range(len(self.overlap_indices) - 1):
            # TODO: either raise an exception or prevent this from happening
            # at all
            current_start = self.overlap_indices[i + 1][0]
            current_end = self.overlap_indices[i][1]
            if current_start <= current_end:
                self.warning = 'warning: overlapping overlaps!'
                print self.warning

        self._has_run = True
        return assembly_dict

    def primers(self, tm=60):
        '''Design primers for amplifying the assembled sequence.

        :param tm: melting temperature (lower than overlaps is best).
        :type tm: float
        :returns: Primer list (the output of coral.design.primers).
        :rtype: list

        '''
        self.primers = coral.design.primers(self.template, tm=tm)
        return self.primers

    def write(self, path):
        '''Write assembly oligos and (if applicable) primers to csv.

        :param path: path to csv file, including .csv extension.
        :type path: str

        '''
        with open(path, 'wb') as oligo_file:
            oligo_writer = csv.writer(oligo_file, delimiter=',',
                                      quoting=csv.QUOTE_MINIMAL)
            oligo_writer.writerow(['name', 'oligo', 'notes'])
            for i, oligo in enumerate(self.oligos):
                name = 'oligo {}'.format(i + 1)
                oligo_len = len(oligo)
                if i != len(self.oligos) - 1:
                    oligo_tm = self.overlap_tms[i]
                    notes = 'oligo length: {}, '.format(oligo_len) + \
                            'overlap Tm: {:.2f}'.format(oligo_tm)
                else:
                    notes = 'oligo length: {}'.format(oligo_len)
                oligo_writer.writerow([name, oligo, notes])
            if self.primers:
                for i, (primer, melt) in enumerate(self.primers):
                    oligo_writer.writerow(['primer {}'.format(i + 1),
                                          primer,
                                          'Tm: {:.2f}'.format(melt)])

    def write_map(self, path):
        '''Write genbank map that highlights overlaps.

        :param path: full path to .gb file to write.
        :type path: str

        '''
        starts = [index[0] for index in self.overlap_indices]
        features = []
        for i, start in enumerate(starts):
            stop = start + len(self.overlaps[i])
            name = 'overlap {}'.format(i + 1)
            feature_type = 'misc'
            strand = 0
            features.append(coral.Feature(name, start, stop, feature_type,
                                          strand=strand))
        seq_map = coral.DNA(self.template, features=features)
        coral.seqio.write_dna(seq_map, path)

    def __repr__(self):
        '''Representation of an OligoAssembly object.'''
        if self._has_run:
            str1 = 'An OligoAssembly consisting of '
            str2 = str(len(self.oligos)) + ' oligos.'
            return str1 + str2
        else:
            return 'An OligoAssembly instance that has not been run.'


def _grow_overlaps(dna, melting_temp, require_even, length_max, overlap_min,
                   min_exception):
    '''Grows equidistant overlaps until they meet specified constraints.

    :param dna: Input sequence.
    :type dna: coral.DNA
    :param melting_temp: Ideal Tm of the overlaps, in degrees C.
    :type melting_temp: float
    :param require_even: Require that the number of oligonucleotides is even.
    :type require_even: bool
    :param length_max: Maximum oligo size (e.g. 60bp price point cutoff)
                       range.
    :type length_range: int
    :param overlap_min: Minimum overlap size.
    :type overlap_min: int
    :param min_exception: In order to meet melting_temp and overlap_min
                          settings, allow overlaps less than overlap_min to
                          continue growing above melting_temp.
    :type min_exception: bool
    :returns: Oligos, their overlapping regions, overlap Tms, and overlap
              indices.
    :rtype: tuple

    '''
    # TODO: prevent growing overlaps from bumping into each other -
    # should halt when it happens, give warning, let user decide if they still
    # want the current construct
    # Another option would be to start over, moving the starting positions
    # near the problem region a little farther from each other - this would
    # put the AT-rich region in the middle of the spanning oligo

    # Try bare minimum number of oligos
    oligo_n = len(dna) // length_max + 1

    # Adjust number of oligos if even number required
    if require_even:
        oligo_increment = 2
        if oligo_n % 2 == 1:
            oligo_n += 1
    else:
        oligo_increment = 1

    # Increase oligo number until the minimum oligo_len is less than length_max
    while float(len(dna)) / oligo_n > length_max:
        oligo_n += oligo_increment

    # Loop until all overlaps meet minimum Tm and length
    tm_met = False
    len_met = False

    while(not tm_met or not len_met):
        # Calculate initial number of overlaps
        overlap_n = oligo_n - 1

        # Place overlaps approximately equidistant over sequence length
        overlap_interval = float(len(dna)) / oligo_n
        starts = [int(overlap_interval * (i + 1)) for i in range(overlap_n)]
        ends = [index + 1 for index in starts]

        # Fencepost for while loop
        # Initial overlaps (1 base) and their tms
        overlaps = [dna[start:end] for start, end in zip(starts, ends)]
        overlap_tms = [coral.analysis.tm(overlap) for overlap in overlaps]
        index = overlap_tms.index(min(overlap_tms))
        # Initial oligos - includes the 1 base overlaps.
        # All the oligos are in the same direction - reverse
        # complementation of every other one happens later
        oligo_starts = [0] + starts
        oligo_ends = ends + [len(dna)]
        oligo_indices = [oligo_starts, oligo_ends]

        oligos = [dna[start:end] for start, end in zip(*oligo_indices)]

        # Oligo won't be maxed in first pass. tm_met and len_met will be false
        maxed = False

        while not (tm_met and len_met) and not maxed:
            # Recalculate overlaps and their Tms
            overlaps = _recalculate_overlaps(dna, overlaps, oligo_indices)
            # Tm calculation is bottleneck - only recalculate changed overlap
            overlap_tms[index] = coral.analysis.tm(overlaps[index])
            # Find lowest-Tm overlap and its index.
            index = overlap_tms.index(min(overlap_tms))
            # Move overlap at that index
            oligos = _expand_overlap(dna, oligo_indices, index, oligos,
                                     length_max)
            # Regenerate conditions
            maxed = any([len(x) == length_max for x in oligos])
            tm_met = all([x >= melting_temp for x in overlap_tms])
            if min_exception:
                len_met = True
            else:
                len_met = all([len(x) >= overlap_min for x in overlaps])

        # TODO: add test for min_exception case (use rob's sequence from
        # 20130624 with 65C Tm)
        if min_exception:
            len_met = all([len(x) >= overlap_min for x in overlaps])

            # See if len_met is true - if so do nothing
            if len_met:
                break
            else:
                while not len_met and not maxed:
                    # Recalculate overlaps and their Tms
                    overlaps = _recalculate_overlaps(dna, overlaps,
                                                     oligo_indices)
                    # Overlap to increase is the shortest one
                    overlap_lens = [len(overlap) for overlap in overlaps]
                    index = overlap_lens.index(min(overlap_lens))
                    # Increase left or right oligo
                    oligos = _expand_overlap(dna, oligo_indices, index, oligos,
                                             length_max)
                    # Recalculate conditions
                    maxed = any([len(x) == length_max for x in oligos])
                    len_met = all([len(x) >= overlap_min for x in overlaps])

                # Recalculate tms to reflect any changes (some are redundant)
                overlap_tms[index] = coral.analysis.tm(overlaps[index])

                # Outcome could be that len_met happened *or* maxed out
                # length of one of the oligos. If len_met happened, should be
                # done so long as tm_met has been satisfied. If maxed happened,
                # len_met will not have been met, even if tm_met is satisfied,
                # and script will reattempt with more oligos

        oligo_n += oligo_increment

    # Calculate location of overlaps
    overlap_indices = [(oligo_indices[0][x + 1], oligo_indices[1][x]) for x in
                       range(overlap_n)]

    return oligos, overlaps, overlap_tms, overlap_indices


def _recalculate_overlaps(dna, overlaps, oligo_indices):
    '''Recalculate overlap sequences based on the current overlap indices.

    :param dna: Sequence being split into oligos.
    :type dna: coral.DNA
    :param overlaps: Current overlaps - a list of DNA sequences.
    :type overlaps: coral.DNA list
    :param oligo_indices: List of oligo indices (starts and stops).
    :type oligo_indices: list
    :returns: Overlap sequences.
    :rtype: coral.DNA list

    '''
    for i, overlap in enumerate(overlaps):
        first_index = oligo_indices[0][i + 1]
        second_index = oligo_indices[1][i]
        overlap = dna[first_index:second_index]
        overlaps[i] = overlap

    return overlaps


def _expand_overlap(dna, oligo_indices, index, oligos, length_max):
    '''Given an overlap to increase, increases smaller oligo.

    :param dna: Sequence being split into oligos.
    :type dna: coral.DNA
    :param oligo_indices: index of oligo starts and stops
    :type oligo_indices: list
    :param index: index of the oligo
    :type index: int
    :param left_len: length of left oligo
    :type left_len: int
    :param right_len: length of right oligo
    :type right_len: int
    :param length_max: length ceiling
    :type length_max: int
    :returns: New oligo list with one expanded.
    :rtype: list

    '''
    left_len = len(oligos[index])
    right_len = len(oligos[index + 1])

    # If one of the oligos is max size, increase the other one
    if right_len == length_max:
        oligo_indices[1] = _adjust_overlap(oligo_indices[1], index, 'right')
    elif left_len == length_max:
        oligo_indices[0] = _adjust_overlap(oligo_indices[0], index, 'left')
    else:
        if left_len > right_len:
            oligo_indices[0] = _adjust_overlap(oligo_indices[0], index, 'left')
        else:
            oligo_indices[1] = _adjust_overlap(oligo_indices[1], index,
                                               'right')

    # Recalculate oligos from start and end indices
    oligos = [dna[start:end] for start, end in zip(*oligo_indices)]

    return oligos


def _adjust_overlap(positions_list, index, direction):
    '''Increase overlap to the right or left of an index.

    :param positions_list: list of overlap positions
    :type positions_list: list
    :param index: index of the overlap to increase.
    :type index: int
    :param direction: which side of the overlap to increase - left or right.
    :type direction: str
    :returns: A list of overlap positions (2-element lists)
    :rtype: list
    :raises: ValueError if direction isn't \'left\' or \'right\'.

    '''
    if direction == 'left':
        positions_list[index + 1] -= 1
    elif direction == 'right':
        positions_list[index] += 1
    else:
        raise ValueError('direction must be \'left\' or \'right\'.')
    return positions_list



'''The central dogma of biology - transcription and translation.'''
import coral
from . import utils


def transcribe(dna):
    '''Transcribe DNA to RNA (no post-transcriptional processing).

    :param seq: Sequence to transcribe (DNA).
    :type seq: coral.DNA
    :returns: Transcribed sequence - an RNA sequence.
    :rtype: coral.RNA

    '''
    return utils.convert_sequence(dna, 'rna')


def translate(rna):
    '''Translate RNA to peptide.

    :param rna: Sequence to translate (RNA).
    :type rna: coral.RNA
    :returns: Translated sequence - a peptide.
    :rtype: coral.Peptide

    '''
    return utils.convert_sequence(rna, 'peptide')


def reverse_transcribe(rna):
    '''Reverse transcribe RNA to DNA.

    :param rna: Sequence to reverse transcribe (RNA).
    :type rna: coral.RNA
    :returns: Reverse-transcribed sequence - a DNA sequence.
    :rtype: coral.DNA

    '''
    return utils.convert_sequence(rna, 'dna')


def coding_sequence(rna):
    '''Extract coding sequence from an RNA template.

    :param seq: Sequence from which to extract a coding sequence.
    :type seq: coral.RNA
    :param material: Type of sequence ('dna' or 'rna')
    :type material: str
    :returns: The first coding sequence (start codon -> stop codon) matched
              from 5' to 3'.
    :rtype: coral.RNA
    :raises: ValueError if rna argument has no start codon.
             ValueError if rna argument has no stop codon in-frame with the
             first start codon.

    '''
    if isinstance(rna, coral.DNA):
        rna = transcribe(rna)
    codons_left = len(rna) // 3
    start_codon = coral.RNA('aug')
    stop_codons = [coral.RNA('uag'), coral.RNA('uga'), coral.RNA('uaa')]
    start = None
    stop = None
    valid = [None, None]
    index = 0
    while codons_left:
        codon = rna[index:index + 3]
        if valid[0] is None:
            if codon in start_codon:
                start = index
                valid[0] = True
        else:
            if codon in stop_codons:
                stop = index + 3
                valid[1] = True
                break
        index += 3
        codons_left -= 1

    if valid[0] is None:
        raise ValueError('Sequence has no start codon.')
    elif stop is None:
        raise ValueError('Sequence has no stop codon.')
    coding_rna = rna[start:stop]

    return coding_rna



'''Reactions for simulating and designing cloning reactions and assemblies.'''
from ._resect import three_resect
from ._resect import five_resect
from ._central_dogma import transcribe
from ._central_dogma import translate
from ._central_dogma import reverse_transcribe
from ._central_dogma import coding_sequence
from ._restriction import digest
from ._pcr import pcr
from ._gibson import gibson
from ._oligo_assembly import assemble_oligos
from ._oligo_assembly import bind_unique



'''Sanger sequencing alignment tools.'''
import coral.analysis

# FIXME: sequencing that goes past 'end' of a circular reference
# is reported as an insertion
# TODO: consensus / master sequence for plotting / report / analysis


class Sanger(object):
    '''Align and analyze Sanger sequencing results.'''
    def __init__(self, reference, results, material='DNA'):
        '''
        :param reference: Reference sequence.
        :type reference: :class:`coral.DNA`
        :param results: Sequencing result string. A list of DNA objects is also
                        valid.
        :type results: list of coral.DNA sequences
        :param material: Material of the alignment sequences - 'DNA' or
                         'Peptide'.
        :type material: str
        :returns: instance of coral.analysis.Sanger (contains alignment and
                  provides analysis/visualization methods

        '''
        # Alignment params / thresholds
        self._gap_open = -25
        self._gap_extend = 0
        self._score_threshold = 100

        # Only one input? Put in list so it can be processed the same.
        if type(results) != list:
            results = [results]
        # Sequences and calculations that get reused
        self._reference = reference
        self._results_input = results
        self._processed = [self._remove_n(x) for x in results]
        self.names = [seq.name for seq in results]

        self._material = material
        if self._material == 'DNA':
            self._matrix = 'DNA_simple'
        elif self._material == 'Peptide':
            self._matrix = 'BLOSUM62'
        else:
            raise ValueError('material arugment must be DNA or Peptide')

        # Align
        print '(Aligning...)'
        self.alignments, self.scores = self._align()
        # Sequencing coverage
        self.coverage = self._find_coverage()
        self._differences = self._find_differences()
        self._difference_n = [sum([len(x) for x in difference]) for difference
                              in self._differences]
        self.mismatches, self.insertions, self.deletions = self._differences

    def report(self):
        '''Report deletions, mismatches, and insertions.'''
        difference_names = ['Mismatches', 'Insertions', 'Deletions']
        print
        print 'Summary: '
        print '--------'
        print
        for name, num in zip(difference_names, self._difference_n):
            print '  {}: {}'.format(name, num)
        print

        for name, difference in zip(difference_names, self._differences):
            difference_n = sum([len(x) for x in difference])
            if difference_n:
                print '## {}'.format(name)
                for i, result_difference in enumerate(difference):
                    if result_difference:
                        result_name = self.names[i]
                        print '  {}'.format(result_name)
                        ref_i = self.alignments[i][0]
                        res_i = self.alignments[i][1]
                        for start, end in result_difference:
                            print
                            _sequences_display(ref_i, res_i, start, end)
                            print

    def plot(self, text=True):
        '''Plot visualization of the alignment results using matplotlib.

        :param text: Control whether text is plotted on top of sequence bars.
        :type text: bool

        '''
        try:
            from matplotlib import pyplot
            from matplotlib import cm
        except ImportError:
            raise ImportError('Optional dependency matplotlib not installed.')

        # Calculations
        # Reference bar information
        reference_x = 0
        reference_y = 14.25
        reference_width = len(self.alignments[0][0])
        reference_height = 1

        # Bin the features so they don't overlap when plotted
        features = self._reference.features
        feature_ranges = [(feature.start, feature.stop) for feature in
                          features]
        feature_bins = disjoint_bins(feature_ranges)
        feature_nbin = len(feature_bins)

        # Bin the alignments so they don't overlap when plotted
        alignment_bins = disjoint_bins(self.coverage)

        # Calculate discrepancy coordinates
        discrepancy_coords = [[], [], []]
        differences = [self.mismatches, self.insertions, self.deletions]
        for i, result_bin in enumerate(alignment_bins):
            for index in result_bin:
                for j, discrepancy_type in enumerate(differences):
                    for discrepancy in discrepancy_type[index]:
                        coord_y = i
                        coord_x = (discrepancy[0] + discrepancy[1]) // 2
                        coords = (coord_x, coord_y)
                        discrepancy_coords[j].append(coords)

        # Plotting
        # Plot calculations
        # Controls spacing between bars, size of bars, etc
        size = 10

        # Matplotlib commands
        # Plot a black reference bar at the bottom
        fig = pyplot.figure(figsize=(12, 9), dpi=90)
        sub1 = fig.add_subplot(111)
        sub1.broken_barh([(reference_x, reference_width)],
                         (reference_y, reference_height),
                         facecolors='black', edgecolors='none')

        dotted_x = (0, len(self.alignments[0][0]))
        height = (feature_nbin + 1.5) * size
        dotted_y = (height, height)
        sub1.plot(dotted_x, dotted_y, linestyle='dotted')

        # Plot the reference features on top of the bar
        # Plot the features by bin:
        for i, feature_bin in enumerate(feature_bins):
            for index in feature_bin:
                feature = features[index]
                name = feature.name

                width = feature.stop - feature.start
                height = size - 1
                y_index = (i + 1) * size
                mid = (feature.start + feature.stop) // 2

                pos = float(i) / len(features)
                sub1.broken_barh([(feature.start, width)],
                                 (y_index, height),
                                 facecolors=cm.Set3(pos),
                                 edgecolors='black')
                if text:
                    sub1.text(mid, y_index + size / 2, name, rotation=90,
                              size='smaller')

        # Plot sequencing results by bin
        for i, result_bin in enumerate(alignment_bins):
            for index in result_bin:
                start, stop = self.coverage[index]
                name = self.names[index]

                width = stop - start
                height = size - 1
                y_index = (i + feature_nbin + 2) * size
                text_x = start + (stop - start) // 8

                sub1.broken_barh([(start, width)], (y_index, height),
                                 facecolors='pink', edgecolors='black')
                if text:
                    sub1.text(text_x, y_index + size // 3, wrap_name(name),
                              rotation=0, size='smaller')

        # Plot mismatches, insertions, deletions
        sub1.plot(1000, 25)
        shapes = ['o', 'x', 'v']
        labels = ['mismatch', 'insertion', 'deletion']
        for coords, shape, label in zip(discrepancy_coords, shapes, labels):
            x_coords = [x[0] for x in coords]
            y_coords = [(x[1] + feature_nbin + 2) * size + 2 for x in coords]
            sub1.scatter(x_coords, y_coords, marker=shape, color='k',
                         label=label)
        sub1.legend()

        # Plot labeling, etc
        sub1.set_xlim(0, reference_width)
        sub1.set_xlabel('Base pairs from origin')
        sub1.set_yticks([15, 15 + size * (feature_nbin + 1)])
        sub1.set_yticklabels(['Reference', 'Results'])
        sub1.grid(True)
        sub1.xaxis.grid(False)
        sub1.yaxis.grid(False)

        pyplot.title('Alignment gap summary', fontstyle='normal')
        pyplot.show()

    def _remove_n(self, seq):
        '''Find largest non-N segment.

        :param seq: Sequence that contains Ns to remove
        :type seq: coral.DNA

        '''
        seq_str = str(seq)
        largest = max([x for x in seq_str.split('N')], key=len)
        seq_start = seq_str.index(largest)
        seq_stop = seq_start + len(largest)
        processed = seq[seq_start:seq_stop]
        return processed

    def _align(self):
        '''Align sequences using needle.'''
        # Align
        refs = [self._reference.copy() for x in self._processed]
        if len(self._processed) > 1:
            needle = coral.analysis.needle_multi(refs, self._processed,
                                                 gap_open=self._gap_open,
                                                 gap_extend=self._gap_extend,
                                                 matrix=self._matrix)
        else:
            needle_single = coral.analysis.needle(refs[0], self._processed[0],
                                                  gap_open=self._gap_open,
                                                  gap_extend=self._gap_extend,
                                                  matrix=self._matrix)
            needle = [needle_single]
        # Split into alignments and scores
        # TODO: use zip here to make it simpler
        alignments = [(str(ref), str(res)) for ref, res, score in
                      needle]
        scores = [result[2] for result in needle]
        # If a result scores too low, try reverse complement
        # TODO: Find their indices and use multiprocessing
        # TODO: If score is too low, add warning but don't show in alignment
        failed = []
        # reversed so we can pop bad alignments (not show them in the plot)
        for i, score in reversed(list(enumerate(scores))):
            if score < self._score_threshold:
                swapped_result = self._processed[i].reverse_complement()
                new_needle = coral.analysis.needle(self._reference,
                                                   swapped_result,
                                                   gap_open=self._gap_open,
                                                   gap_extend=self._gap_extend,
                                                   matrix=self._matrix)
                alignments[i] = (str(new_needle[0]), str(new_needle[1]))
                score = new_needle[2]
            if score < self._score_threshold:
                failed.append(self._processed[i].name)
                # Remove the failed alignment!
                alignments.pop(i)
        if failed:
            msg = 'The following results fell below the score threshold: '
            print msg + '{}'.format(failed)
        # Trim to reference - if reference is shorter than results
        # (e.g. reference is just plasmid insert, results start earlier)
        for i, (reference, result) in enumerate(alignments):
            f_count = 0
            r_count = 0
            # Count beginning gaps (hyphens), then trim
            for char in reference:
                if char != '-':
                    break
                else:
                    f_count += 1
            for char in reversed(reference):
                if char != '-':
                    break
                else:
                    r_count += 1

            if r_count:
                trimmed_ref = reference[f_count:]
                trimmed_res = result[f_count:]
            else:
                trimmed_ref = reference[f_count:]
                trimmed_res = result[f_count:]
            alignments[i] = (trimmed_ref, trimmed_res)

        return alignments, scores

    def _find_coverage(self):
        '''Find sequencing coverage using trailing and leading gaps.'''
        left_coverage = [len(res) - len(res.lstrip('-')) for ref, res in
                         self.alignments]
        right_coverage = [len(res.rstrip('-')) for ref, res in self.alignments]
        return zip(left_coverage, right_coverage)

    def _find_differences(self):
        '''Calculate mismatches, insertions, and deletions.'''
        all_mismatches = [[] for i in range(len(self.alignments))]
        all_insertions = [[] for i in range(len(self.alignments))]
        all_deletions = [[] for i in range(len(self.alignments))]
        # Got through each alignment
        for i, (reference, result) in enumerate(self.alignments):
            # Go through each alignment base by base
            coverage = self.coverage[i]
            ref_trim = reference[slice(*coverage)]
            res_trim = result[slice(*coverage)]
            for j, (refbase, resbase) in enumerate(zip(ref_trim, res_trim)):
                # If only the reference base is a '-', is insertion
                position = j + coverage[0]
                if refbase == '-':
                    all_insertions[i].append((position, position))
                # If only the reference base is a '-', is deletion
                elif resbase == '-':
                    all_deletions[i].append((position, position))
                # If bases don't match, is mismatch
                elif refbase != resbase:
                    all_mismatches[i].append((position, position))

        # Group mismatches/indels if they appear one after another
        all_mismatches = [_group_differences(mismatches) for mismatches in
                          all_mismatches]
        all_insertions = [_group_differences(insertions) for insertions in
                          all_insertions]
        all_deletions = [_group_differences(deletions) for deletions in
                         all_deletions]
        return all_mismatches, all_insertions, all_deletions

    def __repr__(self):
        '''Representation of a Sanger sequencing object.'''
        disc_sum = sum(self._difference_n)
        head = 'An alignment with {} discrepancies.'.format(disc_sum)
        if disc_sum:
            mismatches = 'mismatches: {}, '.format(self._difference_n[0])
            insertions = 'insertions: {}, '.format(self._difference_n[1])
            deletions = 'deletions: {}'.format(self._difference_n[2])
            diffs = '\n' + mismatches + insertions + deletions
        else:
            diffs = ''
        return head + diffs


def _group_differences(difference_list):
    '''Group adjoining insertions or deletions, respectively.

    :param difference_list: list of difference indices (2-tuple).
    :type difference_list: list
    :returns: a list of the same format as the input, but with adjoining
              insertions or deletions grouped together

    '''
    # By setting to (-2, -2), skips first base so that others can
    # 'look back' to see whether the current deletion is last + 1
    previous = (-2, -2)
    grouped_differences = []
    for difference in difference_list:
        if difference[0] - previous[1] == 1:
            grouped_differences[-1] = (previous[0], difference[1])
        else:
            grouped_differences.append(difference)
        previous = grouped_differences[-1]
    return grouped_differences


def _sequences_display(seq1, seq2, start, stop, context=10):
    '''Display two sequences, highlighting the regions where they differ.

    :param seq1: First sequence to compare.
    :type seq1: str
    :param seq2: Second sequence to compare.
    :type seq2: str
    :param start: Where to start displaying the sequence.
    :type start: int
    :param stop: Where to stop displaying the sequence.
    :type stop: int
    :param context: Extra context to add on either side of the displayed
                    sequences.
    :type context: int

    '''
    # TODO: if seqs aren't the same size, should display a little differently
    # TODO: display is incomplete - doesn't show all mismatches / deletions /
    # insertions at once, just the ones displayed at the moment.
    # Should instead calculate a single 'overview' set of sequences:
    # top, bottom, and middle, where middle is where all the seqs match.
    # Then just subset this
    indent = 4
    # Figure out how much context can be included (seq might start/end earlier)
    l_context = (max(start - context - 1, 0), start - 1)
    r_context = (stop + 1, min(start + context + 1, len(seq1)))

    def gen_levels(seq1, seq2, context_tuple):
        '''Generate a text column: left side or right side of difference.

        :param seq1: same as seq1 of parent
        :type seq1: str
        :param seq2: same as seq2 of parent
        :type seq2: str
        :context_tuple: l_context or r_context
        :type context_tuple: tuple

        '''
        top, bottom = [seq[slice(*context_tuple)] for seq in [seq1, seq2]]
        context_len = (context_tuple[1] - context_tuple[0])
        middle = '|' * context_len
        highlight = ' ' * context_len
        return top, bottom, middle, highlight

    # Generate left and right columns of text to display
    l_top, l_bottom, l_middle, l_highlight = gen_levels(seq1, seq2, l_context)
    r_top, r_bottom, r_middle, r_highlight = gen_levels(seq1, seq2, r_context)

    # Generate core column - core is where differences are
    core_slice = slice(start, stop + 1)
    core_len = stop - start + 1
    core_top = seq1[core_slice]
    core_bottom = seq2[core_slice]
    core_middle = ' ' * core_len
    core_highlight = '*' * core_len

    # Combine each level together for printing
    top = l_top + core_top + r_top
    middle = l_middle + core_middle + r_middle
    bottom = l_bottom + core_bottom + r_bottom
    highlight = l_highlight + core_highlight + r_highlight

    indent = ' ' * indent
    if start != stop:
        print '{0}Positions {1} to {2}:'.format(indent, start, stop)
    else:
        print '{0}Position {1}:'.format(indent, start)

    print indent + top
    print indent + middle
    print indent + bottom
    print indent + highlight


def disjoint_bins(ranges_list):
    '''Construct disjoint bins given a list of 1-D ranges (tuples).

    :param range_tuple_list: A list of tuples containing range values.
    :type range_tuple_list: list
    :returns: a list of bins containing indices of the input list. e.g. if
              the third (index 2) range is in the 2nd bin, the number 2 is in
              the 2nd list.
    :rtype: list of lists of ints

    '''
    # Keep track of the original order for reporting later
    ranges_list = [(x[0], x[1], i) for i, x in enumerate(ranges_list)]
    ranges_list = sorted(ranges_list, key=lambda starts: starts[0])

    remaining = ranges_list[:]
    binned = []
    while True:
        current_bin = []
        next_bin = []
        while remaining:
            last = remaining.pop(0)
            current_bin.append(last)
            next_bin += [x for x in remaining if x[0] < last[1]]
            remaining = [x for x in remaining if x[0] >= last[1]]
        binned.append(current_bin)

        if not remaining and not next_bin:
            break
        else:
            remaining = next_bin

    bins = [[x[2] for x in range_bin] for range_bin in binned]

    return bins


def wrap_name(str_in, wrap_len=15):
    '''Wrap plotted text to avoid overlaps (not perfect).

    :param str_in: Input string.
    :type str_in: str
    :param wrap_len: Length at which to wrap lines.
    :type wrap_len: str
    :returns: Wrapped text
    :rtype: str

    '''
    wrap_positions = range(0, len(str_in), wrap_len)
    out = [str_in[i:i + wrap_len] for i in wrap_positions]
    return '\n'.join(out)



'''Resection (need a new name!) - exonuclease activity.'''
import coral


def five_resect(dna, n_bases):
    '''Remove bases from 5' end of top strand.

    :param dna: Sequence to resect.
    :type dna: coral.DNA
    :param n_bases: Number of bases cut back.
    :type n_bases: int
    :returns: DNA sequence resected at the 5' end by n_bases.
    :rtype: coral.DNA

    '''
    new_instance = dna.copy()
    if n_bases >= len(dna):
        new_instance._top.seq = ''.join(['-' for i in range(len(dna))])
    else:
        new_instance._top.seq = '-' * n_bases + str(dna)[n_bases:]

    new_instance = _remove_end_gaps(new_instance)

    return new_instance


def three_resect(dna, n_bases):
    '''Remove bases from 3' end of top strand.

    :param dna: Sequence to resect.
    :type dna: coral.DNA
    :param n_bases: Number of bases cut back.
    :type n_bases: int
    :returns: DNA sequence resected at the 3' end by n_bases.
    :rtype: coral.DNA

    '''
    new_instance = dna.copy()
    if n_bases >= len(dna):
        new_instance._top.seq = ''.join(['-' for i in range(len(dna))])
    else:
        new_instance._top.seq = str(dna)[:-n_bases] + '-' * n_bases

    new_instance = _remove_end_gaps(new_instance)

    return new_instance


def _remove_end_gaps(sequence):
    '''Removes double-stranded gaps from ends of the sequence.

    :returns: The current sequence with terminal double-strand gaps ('-')
              removed.
    :rtype: coral.DNA

    '''
    # Count terminal blank sequences
    def count_end_gaps(seq):
        gap = coral.DNA('-')
        count = 0
        for base in seq:
            if base == gap:
                count += 1
            else:
                break

        return count

    top_left = count_end_gaps(sequence.top())
    top_right = count_end_gaps(reversed(sequence.top()))
    bottom_left = count_end_gaps(reversed(sequence.bottom()))
    bottom_right = count_end_gaps(sequence.bottom())

    # Trim sequence
    left_index = min(top_left, bottom_left)
    right_index = len(sequence) - min(top_right, bottom_right)

    return sequence[left_index:right_index]



# -*- coding: utf-8
'''Vienna RNA module.'''
from subprocess import Popen, PIPE, STDOUT
from tempfile import mkdtemp
from os.path import isdir
from shutil import rmtree


class Vienna(object):
    '''Run Vienna RNA functions on a sequence.'''
    def __init__(self, seqs, dotbrackets=None, constraint_structures=None):
        '''
        :param seq: DNA or RNA sequences to evaluate.
        :type seq: list of coral.DNA, coral.RNA, or str
        :param dotbrackets: Dot bracket formatted structures of given seuqneces
        :type dotbrackets: list of str

        :returns: coral.analysis.Vienna instance.

        '''
        self._seqs = [str(seq) for seq in seqs]
        self._tempdir = ''
        if dotbrackets is None:
            self.dotbrackets = []
        else:
            self.dotbrackets = dotbrackets
        if constraint_structures is None:
            self.constraint_structures = []
        else:
            self.constraint_structures = constraint_structures

    def free_energy_of_structure(self, temp=30.0, index=None):
        '''Calculate the free energy(ies) of sequence(s) and structure(s)
        at a given temperature.

        :param temp: Temperature at which to run calculations (°C).
        :type temp: float
        :param index: Index of the sequence for which to calculate mfe.
        :type index: int
        :returns: Minimum Free Energy (mfe) list [kcal/mol].
        :rtype: list of floats

        '''
        # FIXME: will error if not initialized with dotbrackets
        def run_rnaeval(seq, dotbracket):
            process = Popen(['RNAeval', '-T', str(temp)], stdin=PIPE,
                            stdout=PIPE, stderr=STDOUT)
            rnaeval_input = seq + '\n' + dotbracket
            output = process.communicate(input=rnaeval_input)[0]
            lines = output.splitlines()
            lines = lines[-1].split('(')[-1].split(')')[0].strip()

            return float(lines)

        mfes = []
        if index is None:
            for seq, dotbracket in zip(self._seqs, self.dotbrackets):
                mfes.append(run_rnaeval(seq, dotbracket))
        else:
            for seq, dotbracket in zip(self._seqs[index],
                                       self.dotbrackets[index]):
                mfes.append(run_rnaeval(seq, dotbracket))
        return mfes

    def mfe(self, temp=50.0, return_structure=False, index=None,
            use_constraint_structures=False):
        '''Calculate the minimum free energy.

        :param temp: Temperature at which to run calculations.
        :type temp: float
        :param index: Specific sequence index on which to calcualte mfe.
        :type index: int
        :returns: Minimum Free Energy (mfe) list [kcal/mol].
        :rtype: list of floats

        '''
        mfes = []
        dotbrackets = []

        def run_rnafold(seq, constraint=''):
            arguments = ['RNAfold', '-T', str(temp), '-noPS', '-C']

            process = Popen(arguments, stdin=PIPE,
                            stdout=PIPE, stderr=STDOUT)
            rnafold_input = '\n'.join([seq, constraint])
            output = process.communicate(input=rnafold_input)[0]
            lines = output.splitlines()
            mfe = lines[-1].split('(')[-1].split(')')[0].strip()
            dotbracket = lines[-1].split(' ')[0].strip()

            return mfe, dotbracket

        if use_constraint_structures:
            if index is None:
                for seq, const in zip(self._seqs, self.constraint_structures):
                    mfe, dotbracket = run_rnafold(seq, constraint=const)
                    mfes.append(float(mfe))
                    dotbrackets.append(dotbracket)
            else:
                const = self.sequence_constraints[index]
                mfe, dotbracket = run_rnafold(self._seqs[index],
                                              constraint=const)
                mfes.append(float(mfe))
                dotbrackets.append(dotbracket)
        else:
            if index is None:
                for seq in self._seqs:
                    mfe, dotbracket = run_rnafold(seq)
                    mfes.append(float(mfe))
                    dotbrackets.append(dotbracket)
            else:
                mfe, dotbracket = run_rnafold(self._seqs[index])
                mfes.append(float(mfe))
                dotbrackets.append(dotbracket)

        if return_structure:
            return zip(mfes, dotbrackets)
        else:
            return mfes

    def unbound_probability(self, temp=50.0, index=None):
        '''Calculate per-pair probability of being unbound (secondary
        structure).

        :param temp: Temperature at which to run calculations.
        :type temp: float
        :param index: Specific sequence index on which to calcualte pairs.
        :type index: int
        :returns: Pair probability for every base in the sequences.
        :rtype: list of list of floats.

        '''
        def run_rnafold(seq):
            process = Popen(['RNAfold', '-p', '-T', str(temp)], stdin=PIPE,
                            stdout=PIPE, stderr=STDOUT, cwd=self._tempdir)
            process.communicate(input=seq)[0]
            with open('{}/dot.ps'.format(self._tempdir), 'r') as dot:
                text = dot.read()
                text = text[text.index('%data'):]
            split = text.split('\n')
            data = [x for x in split if x.endswith('ubox')]
            data = [x.rstrip(' ubox') for x in data]
            data = [x.split() for x in data]
            data = [(int(a), int(b), float(c)) for a, b, c in data]
            unbound = [1.0] * len(seq)
            for base1, base2, prob_sqr in data:
                probability = prob_sqr**2
                unbound[base1 - 1] -= probability
                unbound[base2 - 1] -= probability
            return unbound

        self._check_tempdir()
        unbounds = []
        if index is None:
            for seq in self._seqs:
                unbound = run_rnafold(seq)
                unbounds.append(unbound)
        else:
            for seq in self._seqs[index]:
                unbound = run_rnafold(seq)
                unbounds.append(unbound)

        self._close()
        return unbounds

    def hybridization_mfe(self, temp=37.0, simtype='duplex', indices=None):
        '''Calculate the mfe of two RNA seqeunces complexing.

        :param temp: Temperature at which to run calculations.
        :type temp: float
        :param indices: list (len 2) of sequence indices on which to
                        calculate pairs.
        :type indices: list of ints
        :param simtype: Vienna simulation package to use. \'duplex\' dose not
                        allow self RNA interactions. \'cofold\' allows self RNA
                        interactions to occur.
        :type simtype: str
        :returns: Binding energy between two given sequence
        :rtype: (MFE, [dotbracket1, dotbracket2], [[start1,stop1],
                [[start2,stop2]])
        :rtype: list of list of floats.

        '''
        # TODO: refactor
        if indices is None:
            tempseqs = self._seqs
        else:
            tempseqs = [self._seqs[indices[0]], self._seqs[indices[1]]]
        if simtype == 'cofold':
            process = Popen(['RNAcofold', '-T', str(temp)], stdin=PIPE,
                            stdout=PIPE, stderr=STDOUT, cwd=self._tempdir)
            stringtoinput = tempseqs[0] + '&' + tempseqs[1]
            output = process.communicate(input=stringtoinput)[0]
            lines = output.splitlines()
            mfe = float(lines[-1].split('(')[-1].split(')')[0].strip())
            # TODO: make these produce calculated values
            brackets = []
            bindlocations = []
        elif simtype == 'duplex':
            process = Popen(['RNAduplex', '-T', str(temp)], stdin=PIPE,
                            stdout=PIPE, stderr=STDOUT, cwd=self._tempdir)
            stringtoinput = tempseqs[0] + '\n' + tempseqs[1]
            output = process.communicate(input=stringtoinput)[0]
            mfe = float(output.split('(')[-1].split(')')[0])
            brackets = [output.split('&')[0],
                        output.split('&')[1].split(' ')[0]]
            bind_location_1 = [int(i) for i in
                               output.split('  ')[1].strip().split(',')]
            bind_location_2 = [int(i) for i in
                               output.split('  ')[3].strip().split(',')]
            bindlocations = [bind_location_1, bind_location_2]
        return [(mfe, brackets, bindlocations)]

    def _close(self):
        '''Close the temporary dir (keeps /tmp clean).'''
        rmtree(self._tempdir)

    def _check_tempdir(self):
        '''If temp dir has been removed, create a new one.'''
        if not isdir(self._tempdir):
            self._tempdir = mkdtemp()



'''Restriction endonuclease reactions.'''
import coral.reaction


def digest(dna, restriction_enzyme):
    '''Restriction endonuclease reaction.

    :param dna: DNA template to digest.
    :type dna: coral.DNA
    :param restriction_site: Restriction site to use.
    :type restriction_site: RestrictionSite
    :returns: list of digested DNA fragments.
    :rtype: coral.DNA list

    '''
    pattern = restriction_enzyme.recognition_site
    located = dna.locate(pattern)
    if not located[0] and not located[1]:
        return [dna]
    # Bottom strand indices are relative to the bottom strand 5' end.
    # Convert to same type as top strand
    pattern_len = len(pattern)
    r_indices = [len(dna) - index - pattern_len for index in
                 located[1]]
    # If sequence is palindrome, remove redundant results
    if pattern.is_palindrome():
        r_indices = [index for index in r_indices if index not in
                     located[0]]
    # Flatten cut site indices
    cut_sites = sorted(located[0] + r_indices)
    # Go through each cut site starting at highest one
    # Cut remaining template once, generating remaining + new
    current = [dna]
    for cut_site in cut_sites[::-1]:
        new = _cut(current, cut_site, restriction_enzyme)
        current.append(new[1])
        current.append(new[0])
    current.reverse()
    # Combine first and last back together if digest was circular
    if dna.topology == 'circular':
        current[0] = current.pop() + current[0]
    return current


def _cut(dna, index, restriction_enzyme):
    '''Cuts template once at the specified index.

    :param dna: DNA to cut
    :type dna: coral.DNA
    :param index: index at which to cut
    :type index: int
    :param restriction_enzyme: Enzyme with which to cut
    :type restriction_enzyme: coral.RestrictionSite
    :returns: 2-element list of digested sequence, including any overhangs.
    :rtype: list

    '''
    # TODO: handle case where cut site is outside of recognition sequence,
    # for both circular and linear cases where site is at index 0
    # Find absolute indices at which to cut
    cut_site = restriction_enzyme.cut_site
    top_cut = index + cut_site[0]
    bottom_cut = index + cut_site[1]

    # Isolate left and ride sequences
    to_cut = dna.pop()
    max_cut = max(top_cut, bottom_cut)
    min_cut = min(top_cut, bottom_cut)
    left = to_cut[:max_cut]
    right = to_cut[min_cut:]

    # If applicable, leave overhangs
    diff = top_cut - bottom_cut
    if not diff:
        # Blunt-end cutter, no adjustment necessary
        pass
    elif diff > 0:
        # 3' overhangs
        left = coral.reaction.five_resect(left.flip(), diff).flip()
        right = coral.reaction.five_resect(right, diff)
    else:
        # 5' overhangs
        left = coral.reaction.three_resect(left, abs(diff))
        right = coral.reaction.three_resect(right.flip(), abs(diff)).flip()

    return [left, right]



'''Check sequences for repeats that may impact cloning efficiency.'''
from collections import Counter


def repeats(seq, size):
    '''Count times that a sequence of a certain size is repeated.

    :param seq: Input sequence.
    :type seq: coral.DNA or coral.RNA
    :param size: Size of the repeat to count.
    :type size: int
    :returns: Occurrences of repeats and how many
    :rtype: tuple of the matched sequence and how many times it occurs

    '''
    seq = str(seq)
    n_mers = [seq[i:i + size] for i in range(len(seq) - size + 1)]
    counted = Counter(n_mers)
    # No one cares about patterns that appear once, so exclude them
    found_repeats = [(key, value) for key, value in counted.iteritems() if
                     value > 1]
    return found_repeats



'''Genbank constants'''

TO_CORAL = {'-10_signal': '-10_signal',
            '-35_signal': '-35_signal',
            '3\'UTR': '3\'UTR',
            '5\'UTR': '5\'UTR',
            'assembly_gap': 'assembly_gap',
            'attenuator': 'attenuator',
            'C_region': 'C_region',
            'CAAT_signal': 'CAAT_signal',
            'CDS': 'CDS',
            'centromere': 'centromere',
            'D-loop': 'D-loop',
            'D_segment': 'D_segment',
            'enhancer': 'enhancer',
            'exon': 'exon',
            'gap': 'gap',
            'GC_signal': 'GC_signal',
            'gene': 'gene',
            'iDNA': 'iDNA',
            'intron': 'intron',
            'J_segment': 'J_segment',
            'LTR': 'LTR',
            'mat_peptide': 'mat_peptide',
            'misc_binding': 'misc_binding',
            'misc_difference': 'misc_difference',
            'misc_feature': 'misc_feature',
            'misc_recomb': 'misc_recomb',
            'misc_RNA': 'misc_RNA',
            'misc_signal': 'misc_signal',
            'misc_structure': 'misc_structure',
            'mobile_element': 'mobile_element',
            'modified_base': 'modified_base',
            'mRNA': 'mRNA',
            'ncRNA': 'ncRNA',
            'N_region': 'N_region',
            'old_sequence': 'old_sequence',
            'operon': 'operon',
            'oritT': 'oriT',
            'polyA_signal': 'polyA_signal',
            'polyA_site': 'polyA_site',
            'precursor_RNA': 'precursor_RNA',
            'prim_transcript': 'prim_transcript',
            'primer_bind': 'primer_bind',
            'promoter': 'promoter',
            'protein_bind': 'protein_bind',
            'RBS': 'RBS',
            'repeat_region': 'repeat_region',
            'rep_origin': 'rep_origin',
            'rRNA': 'rRNA',
            'S_region': 'S_region',
            'sig_peptide': 'sig_peptide',
            'source': 'source',
            'stem_loop': 'stem_loop',
            'STS': 'STS',
            'TATA_signal': 'TATA_signal',
            'telomere': 'telomere',
            'terminator': 'terminator',
            'tmRNA': 'tmRNA',
            'transit_peptide': 'transit_peptide',
            'tRNA': 'tRNA',
            'unsure': 'unsure',
            'V_region': 'V_region',
            'V_segment': 'V_segment',
            'variation': 'variation'}
# 'site' appears in some ApE files / Geneious exports, not sure why
TO_CORAL['site'] = 'site'
TO_BIO = {value: key for key, value in TO_CORAL.iteritems()}



'''Generate a random DNA sequence.'''
import random
import coral
from coral.constants.molecular_bio import CODON_FREQ_BY_AA


def random_dna(n):
    '''Generate a random DNA sequence.

    :param n: Output sequence length.
    :type n: int
    :returns: Random DNA sequence of length n.
    :rtype: coral.DNA

    '''
    return coral.DNA(''.join([random.choice('ATGC') for i in range(n)]))


def random_codons(peptide, frequency_cutoff=0.0, weighted=False, table=None):
    '''Generate randomized codons given a peptide sequence.

    :param peptide: Peptide sequence for which to generate randomized
                    codons.
    :type peptide: coral.Peptide
    :param frequency_cutoff: Relative codon usage cutoff - codons that
                             are rarer will not be used. Frequency is
                             relative to average over all codons for a
                             given amino acid.
    :param frequency_cutoff: Codon frequency table to use.
    :param weighted: Use codon table
    :type weighted: bool
    :param table: Codon frequency table to use. Table should be organized
                  by amino acid, then be a dict of codon: frequency.
                  Only relevant if weighted=True or frequency_cutoff > 0.
                  Tables available:

                  constants.molecular_bio.CODON_FREQ_BY_AA['sc'] (default)
    :type table: dict
    :returns: Randomized sequence of codons (DNA) that code for the input
              peptide.
    :rtype: coral.DNA
    :raises: ValueError if frequency_cutoff is set so high that there are no
             codons available for an amino acid in the input peptide.

    '''
    if table is None:
        table = CODON_FREQ_BY_AA['sc']
    # Process codon table using frequency_cutoff
    new_table = _cutoff(table, frequency_cutoff)
    # Select codons randomly or using weighted distribution
    rna = ''
    for amino_acid in str(peptide):
        codons = new_table[amino_acid.upper()]
        if not codons:
            raise ValueError('No {} codons at freq cutoff'.format(amino_acid))
        if weighted:
            cumsum = []
            running_sum = 0
            for codon, frequency in codons.iteritems():
                running_sum += frequency
                cumsum.append(running_sum)
            random_num = random.uniform(0, max(cumsum))
            for codon, value in zip(codons, cumsum):
                if value > random_num:
                    selection = codon
                    break
        else:
            selection = random.choice(codons.keys())
        rna += selection
    return coral.RNA(rna)


def _cutoff(table, frequency_cutoff):
    '''Generate new codon frequency table given a mean cutoff.

    :param table: codon frequency table of form {amino acid: codon: frequency}
    :type table: dict
    :param frequency_cutoff: value between 0 and 1.0 for mean frequency cutoff
    :type frequency_cutoff: float
    :returns: A codon frequency table with some codons removed.
    :rtype: dict

    '''
    new_table = {}
    # IDEA: cutoff should be relative to most-frequent codon, not average?
    for amino_acid, codons in table.iteritems():
        average_cutoff = frequency_cutoff * sum(codons.values()) / len(codons)
        new_table[amino_acid] = {}
        for codon, frequency in codons.iteritems():
            if frequency > average_cutoff:
                new_table[amino_acid][codon] = frequency
    return new_table
